//PRELOAD IMAGES

function preloadImage(src) {
  var preload1 = new Image()
  preload1.classList.add("hidden")
  document.body.appendChild(preload1)
  preload1.onload = function () {
    preload1.remove()
  }
  preload1.src = src
}

let sageSource = "./Menu/SAGE22LOGO.png"

preloadImage("./Menu/Logo.png")
preloadImage(sageSource)
preloadImage("./Menu/Sawfish Team New.png")
preloadImage("./Menu/BannerBg.png")

//=========================================================
//INITIALIZE Variables

function appendChildren(parent, children) {
  for (var child of children) {
    parent.appendChild(child)
  }
}

function createAlert(text) {
  box = document.createElement("div")
  let code = {
    "\n": "<br>"
  }
  let string = new RegExp('\\b(' + Object.keys (code).join ('|') + ')\\b', 'g')
  text = text.replaceAll(string, function (_, str) { return code[str]; })
      box.innerHTML = `<center>${text}<br><br><i>Click to Continue...</i></center>`
      box.style.fontSize = "20px"
      box.style.padding = "50px"
      box.style.background = "darkred"
      box.style.color = "white"
      document.body.appendChild(box)
  return box
}

function playMapMusic() {
  if (! mapmusic.paused) return
  if (mapmusic.src != levelplan[currentlevel]["mapmusic"]) {
    mapmusic.src = levelplan[currentlevel]["mapmusic"]
  } 
    playMusic(mapmusic)
}

function pauseMapMusic() {
  if (mapmusic.paused) return
  mapmusic.pause()
}

function createtextdesc(testMap, testLevel, DOMDisplay) {
  let textdesc = createAlert(levelplan[currentlevel]["textdesc"])
  textdesc.onclick = function () {
  buttonNoise()
  textdesc.remove()
  runGame([testMap, testLevel], DOMDisplay)
  }
}

let currentlevel = 0
let globalPlayerHP = 5
let globalEnemyHP = 5
let power = 0
let allButtons
//var scale = 20
let tutorialtimer = true
attackDirection = "";

const mapKeys = trackKeys(["ArrowLeft", "ArrowRight", "ArrowUp", "ArrowDown", "e"], ["keydown", "keydown", "keydown", "keydown", "keyup"])
const dialogueKeys = trackKeys([" ", "q"], ["keyup", "keydown"])
const levelKeys = trackKeys(["ArrowLeft", "ArrowRight", "ArrowUp", "ArrowDown", "w", "s", "a", "d"], ["keydown", "keydown", "keyup", "keydown", "keyup", "keyup", "keyup", "keyup"])
var arrowKeys = mapKeys;

const attackTypes = ["U", "D", "L", "R"]
let attackData = [Attack, SineAttack, BounceProjectile]

let jump = new Audio(soundEffectData["jump"])
let attack = new Audio(soundEffectData["attack"])
let damage = new Audio(soundEffectData["damage"])
let doublejump = new Audio(soundEffectData["doublejump"])
let tennisball = new Audio(soundEffectData["tennisball"])
let counterattack = new Audio(soundEffectData["counterattack"])
let maxpower = new Audio(soundEffectData["maxpower"])
let enemybanner = new Audio(soundEffectData["enemybanner"])
let cleardata = new Audio(soundEffectData["cleardata"])
let nodata = new Audio(soundEffectData["nodata"])
let stairssound = new Audio(soundEffectData["stairs"])
let warning = new Audio(soundEffectData["warning"])
let buttonsound = new Audio(soundEffectData["button"])

let bannermusic = new Audio(musicData["banner"])
let bgMusic = new Audio(musicData["level"])
let titlemusic = new Audio(musicData["title"])
let mapmusic = new Audio(levelplan[currentlevel]["mapmusic"])

function buttonNoise() {
  playSoundEffect(buttonsound)
}

//=========================================================
//Data Saving Functions

let localStorage = window.localStorage;
let storage = true;
if (!window.localStorage){
  storage = false;
  alert("Local storage is not supported. If you would like to continue from where you left, please use the Save Codes.")
}

function saveData() {
  globalPlayerHP = 5
  if (!storage) return
  if (!Number.isInteger(currentlevel)) {
    currentlevel = 0
    localStorage.setItem("level", "0")
    return
  } 
  let response = confirm(`You have completed Part ${currentlevel - 1} of the game. The Save Code for this section is ${codes[currentlevel]}. Press "OK" to continue and save the game or press "Cancel" to return to the Main Menu without saving.`)
  if (response) {
    localStorage.setItem("level", `${currentlevel}`)
    console.log("Saved!")
  }
  else {
    location.reload()
  }
}

function clearData() {
  if (!storage) return
  localStorage.clear()
  console.log("Data Cleared!")
}

function getData() {
  if (!storage) return
  storedData = localStorage.getItem("level")
  if (storedData == null) {
    return false
  }
  else {
    if (isNaN(parseInt(storedData)) || parseInt(storedData) > levelplan.length - 1) {
      localStorage.setItem("level", "0")
      return 0
    }
    return parseInt(storedData)
  }
}

function checkCode() {
  let input = prompt("Please enter your Save Code:")
  if (codes.includes(input)) {
    return codes.indexOf(input)
  }
  else {
    console.log("Invalid Code")
    return false
  }
}

//=========================================================
//CREDITS Function

function rollCredits(bg, button) {
  let credits = document.createElement("div")
  credits.id = "credits"
  credits.innerHTML = creditsData.join("<br>")
  bg.appendChild(credits);
  setTimeout(function () {
    credits.remove()
    for (i = 0; i < button.length; i ++) bg.appendChild(button[i])
  }, 40000)
}

//=========================================================
//BEGINS the game!

window.addEventListener('DOMContentLoaded', (event) => {
  
    console.log('DOM fully loaded and parsed');
    let testMap = new Level(levelplan[currentlevel]["map"], "map");
    let testLevel = new Level(levelplan[currentlevel]["level"], "level");
    let startScreen = document.createElement("div")
    startScreen.id = "startscreen"
    let sawfishTeam = document.createElement("div")
    sawfishTeam.id = "sawfishteam"
    let text = document.createElement("div")
    sawfishTeam.appendChild(text)
    text.innerHTML = "Click to Continue..."
    text.style.position = "absolute";
    text.style.width = "100%";
    text.style.bottom = "5%";
  let movingBg = document.createElement("div")
  movingBg.id = "movingBg"
    let codeButton = document.createElement("button")
    codeButton.id = "codebutton"
    codeButton.innerHTML = "Enter Save Code"
    let startButton = document.createElement("button")
    startButton.id = "startbutton"
    let creditsButton = document.createElement("button")
    creditsButton.id = "creditsbutton"
    creditsButton.innerHTML = "Credits"
    startButton.innerHTML = "New Game"
    let continueButton = document.createElement("button")
    continueButton.id = "continuebutton"
    let resetButton = document.createElement("button")
    resetButton.id = "resetbutton"
    continueButton.innerHTML = "Continue"
    resetButton.innerHTML = "Clear Progress"

    let bgwrapper = document.createElement("div")
    bgwrapper.id = "bgWrapper"
    let loading = document.createElement("img")
    loading.src = "./Menu/Icon.png"
    loading.id = "loading"
    bgwrapper.appendChild(loading)

    //Music preload
    titlemusic.addEventListener("canplaythrough", function () {
            console.log("Music Loaded")
            loading.remove()
            appendChildren(startScreen, [startButton, creditsButton, resetButton, continueButton, codeButton])
            })

  
    let loadedImages = 0 //PRELOAD the rest of images
    let preloadImages = walkingSpriteData["Aysel"].concat(textureData, walkingSpriteData["Lindor"], menuData, walkingSpriteData["Petros"], walkingSpriteData["Sporky"], dialoguePreLoad, walkingSpriteData["Rufus"])//ADD DIALOGUE SPRITES
      for (let i = 0; i < preloadImages.length; i++)
      {
        var image = new Image()
        image.classList.add("hidden")
        document.body.appendChild(image)
        image.onload = function () {
          loadedImages ++
          if (loadedImages == preloadImages.length) {
            // let preloads = document.getElementsByClassName("hidden")
            // while (preloads.length > 0) {
            //   preloads[0].remove()
            // }
            console.log("Images Loaded")
            //BUTTONS and MAIN MENU

            allButtons = [startButton, codeButton, continueButton, resetButton]
          }
        }
        image.src = preloadImages[i]
      }

    creditsButton.onclick = function() {
      buttonNoise()
      creditsButton.remove()
      startButton.remove()
      resetButton.remove()
      continueButton.remove()
      codeButton.remove()
      rollCredits(bgwrapper, allButtons)
    }
    resetButton.onclick = function() {
      clearData()
      playSoundEffect(cleardata)
    }
    codeButton.onclick = function() {
      let output = checkCode()
      if (!output) {
        playSoundEffect(nodata)
      }
      else {
        currentlevel = output
        let testMap = new Level(levelplan[currentlevel]["map"], "map"); //used to be map
        let testLevel = new Level(levelplan[currentlevel]["level"], "level");
        bgwrapper.remove()
        titlemusic.pause()
        titlemusic.volume = 0
        createtextdesc(testMap, testLevel, DOMDisplay)
      }

    }
//=========================================================
//CONTINUE with save data
    continueButton.onclick = function() {
      if (!getData()) {
        playSoundEffect(nodata)
      }
      else {
        buttonNoise()
        currentlevel = getData()
        let testMap = new Level(levelplan[currentlevel]["map"], "map"); //used to be map
        let testLevel = new Level(levelplan[currentlevel]["level"], "level");
        bgwrapper.remove()
        titlemusic.pause()
        titlemusic.volume = 0
        createtextdesc(testMap, testLevel, DOMDisplay)
      }
    }
    window.addEventListener("mousedown", showStartScreen)
//=========================================================
//Function to show Main Menu when clicked
    function showStartScreen(event) {
      titlemusic = new Audio(musicData["title"])
      playMusic(titlemusic)
      let SAGE = document.createElement("img")
      SAGE.src = sageSource
      SAGE.id = "sage"
      sawfishTeam.style.filter = "opacity(0%)"
      document.body.appendChild(bgwrapper)
      bgwrapper.appendChild(movingBg)
      bgwrapper.appendChild(startScreen)
      bgwrapper.appendChild(SAGE)
      sawfishTeam.remove()
      window.removeEventListener("mousedown", showStartScreen)
    }
//=========================================================
//STARTS game from Level 0
    startButton.onclick = function() {
      buttonNoise()
      bgwrapper.remove()
      titlemusic.pause()
      titlemusic.volume = 0
      instructionText = createAlert("Here are your basic controls:\n\nARROW KEYS can be used to move around\nE can be used to interact with Dialogue squares\nQ can be used to speed up dialogue\nSPACE can be used to exit or skip dialogue\nWASD can be used for movement and Attack blocking\nA+D makes a Horizontal Counterattack\nW+S makes a Vertical Counterattack\n\nThank you for playing the Headmaster Disaster Demo!\nGood luck!")
      instructionText.onclick = function () {
        buttonNoise()
        instructionText.remove()
        createtextdesc(testMap, testLevel, DOMDisplay)
        
      }
      
    }
    document.body.appendChild(sawfishTeam)
    sawfishTeam.style.filter = "opacity(100%)"
    
});

//=========================================================
//FUNCTION to play music and sound effects
async function playMusic(audio) {
  audio.play()
  audio.addEventListener('ended', function() {
    this.currentTime = 0;
    this.play();
  });
}
async function playSoundEffect(audio) {
  audio.play()
}

//=========================================================
//INITIALIZATION of Classes
class Level { 
  constructor(plan, leveltype) {
    let rows = plan.trim().split("\n").map(l => [...l]);
    this.height = rows.length;
    this.width = rows[0].length;
    this.startActors = [];
    this.leveltype = leveltype;

    this.rows = rows.map((row, y) => {
      return row.map((ch, x) => {
        let type = levelChars[ch]; //text identifier
        if (typeof type == "string") return type;
        
        if (type == Dialogue) {
          console.log("Dialogue Object", x, y)
            this.startActors.push(type.create(new Vec(x, y), ch, levelplan[currentlevel]["dialoguegrid"][x][y]))
        }
        else {
            this.startActors.push(type.create(new Vec(x, y), ch)); //sets type to xy position vector
        }
        return "empty";
      });
    });
  }
  get type() { return this.leveltype; }
}

class State {
  constructor(level, actors, status) {
    this.level = level;
    this.actors = actors;
    this.status = status;
    this.tutorial = "";
  }

  static start(level) {
    return new State(level, level.startActors, "playing");
  }

  get player() {
    return this.actors.find(a => a.type == "player");
  }

  get enemy() {
    return this.actors.find(a => a.type == "enemy")
  }
}

// VEC USED TO BE HERE -> Moved to attacks.js

class Player {
  constructor(pos, speed, frame, dir, doubleJump) {
    this.pos = pos;
    this.speed = speed;
    this.attackDirection = "";
    this.frame = frame;
    this.dir = dir;
    this.doubleJump = doubleJump;
  }

  get type() { return "player"; }

  static create(pos) {
    return new Player(pos.plus(new Vec(0, -1.5)), new Vec(0, 0));
  }
}

Player.prototype.size = new Vec(1, 2.5);

class HP {
  constructor(pos, char) {
    this.pos = pos;
    this.char = char;
    this.value = this.char == "player" ? new Vec(globalPlayerHP,1): new Vec(globalEnemyHP * 5 / levelplan[currentlevel]["enemyHealth"], 1);
    this.shift = 0;
  }

  get type() {return "hp"}

  static create (pos, char) {
    return new HP(pos, char);
  }
}

class PowerLevel {
  constructor(pos) {
    this.pos = pos;
    this.shift = 0;
  }

  get type() {return "powerbar"}

  static create (pos) {
    return new PowerLevel(pos);
  }
}

PowerLevel.prototype.size = new Vec(1, 0);

HP.prototype.size = new Vec(5, 1);

class Enemy {
  constructor(pos, timer) {
    this.pos = pos;
    this.timer = 0;
    this.frame = 0;
    this.direction = 0
    this.speed = new Vec(0, 0)
    globalEnemyHP = 5
    attackData = levelplan[currentlevel]["attacks"]
  }

  get type() {return "enemy";}

  static create(pos) {
    return new Enemy(pos.plus(new Vec(0, -1.5)), new Vec(0, 0));
  }
}

Enemy.prototype.size = new Vec(1, 2.5);

class Event {
  constructor(pos) {
    this.pos = pos;
  }

  static create (pos, ch) {
    return new Event(pos)
  }

  get type() {return "event";}
}

Event.prototype.size = new Vec(1,1)

class Stairs {
  constructor(pos) {
    this.pos = pos
  }
  
  static create (pos, ch) {
    return new Stairs(pos)
  }

  get type() { return "stairs"; }
}

Stairs.prototype.size = new Vec(1, 1)

class Dialogue {
  constructor(pos, data) {
    this.pos = pos;
    this.data = data;
  }

  static create (pos, ch, data) {
    if (ch == "?") {
      return new Dialogue(pos, data);
    }
  }

  get type() { return "dialogue"; }
}

Dialogue.prototype.size = new Vec(1,1);

class CounterAttackH {
  constructor(pos, speed) {
    this.pos = pos;
    this.speed = speed;
  }

  get type() {return "counterattack horizontal"}

  static create(pos, speed) {
    return new CounterAttackH(pos, new Vec(speed, 0))
  }
}

class CounterAttackV {
  constructor(pos, speed) {
    this.pos = pos;
    this.speed = speed;
  }
  
  get type () {return "counterattack vertical"}

  static create(pos, speed) {
    return new CounterAttackV(pos, new Vec(0, speed))
  }
}

CounterAttackV.prototype.size = new Vec(1, 1);

CounterAttackH.prototype.size = new Vec(1, 1);

//ATTACKS USED TO BE HERE -> Moved to attacks.js

//Initialization of levelChars for level strings
const levelChars = {
  ".": "empty", "#": "wall",
  "@": Player,
  "!": Event, "?": Dialogue, "E": Enemy,
  "^": Stairs
};


//=========================================================
//FUNCTIONS for Dialogue displays

function printText(text, x, textbox, div, i, dom, dialoguebox, status, data) {
  if (x < text.length && (status == "reading" || status == "gamewin")) {
  textbox.innerHTML += text[x]
  if (dialogueKeys["q"] && x < text.length - 1) {
    x++
    textbox.innerHTML += text[x]
  }
    //document.body.scrollIntoView(false)
    window.scrollTo({ left: 0, top: document.body.scrollHeight});//automatic scrolling
  setTimeout(function () {printText(text, x+1, textbox, div, i, dom, dialoguebox, status, data)}, 50)
  }
  else {
    printDialogueData(div, i+1, dom, dialoguebox, status, data)
  }
}

function printDialogueData(div, i, dom, dialoguebox, status, data){
  if (i < data.length && (status == "reading" || status == "gamewin")) {
      if (data[i]["character"] == "SoundEffect") {
        switch(data[i]["sound"]) {
          case "enemybanner":
            playSoundEffect(enemybanner)
            break
          case "stairs":
            playSoundEffect(stairssound)
            break
        }
        printDialogueData(div, i+1, dom, dialoguebox, status, data)
      }
    else {
      let container = document.createElement("div")
      container.setAttribute("display", "block")
      container.setAttribute("clear", "both")
      container.classList.add("container")
      dialoguebox.appendChild(container)
      let image = document.createElement("img")
      image.setAttribute("src", characterData[data[i]["character"]][data[i]["expression"]])
      image.style.position = "relative"
      let imagewidth = window.innerWidth / 6
      image.setAttribute("width", `${imagewidth}px`)
      // image.style.left = `${window.innerWidth / 2 - imagewidth}px`
      image.classList.add("dialogueimage")
      container.appendChild(image)
      let text = data[i]["text"]
      let textbox = document.createElement("div")
      textbox.classList.add("dialoguetext")
      textbox.id = `dialoguetext${i}`
      container.appendChild(textbox)
      document.body.insertBefore(dialoguebox, dom)
      printText(text, 0, document.getElementById(`dialoguetext${i}`), div, i, dom, dialoguebox, status, data)
    }
  }
    else if (i >= data.length && status == "gamewin") {
      dialoguebox.remove()
      currentlevel ++
      if (currentlevel < levelplan.length) {
        saveData()
        arrowKeys = mapKeys
        let testMap = new Level(levelplan[currentlevel]["map"], "map"); //used to be map
        let testLevel = new Level(levelplan[currentlevel]["level"], "level");
        createtextdesc(testMap, testLevel, DOMDisplay)
        return
      }
      let endScreen = document.createElement("div")
      let endWrapper = document.createElement("div")
      let endBg = document.createElement("div")
      let restartButton = document.createElement("button")
      restartButton.id = "startbutton"
      restartButton.innerHTML = "Restart"
      restartButton.onclick = function() {
        buttonsound.play()
        location.reload();
      }
      endScreen.id = "startscreen"
      endWrapper.id = "bgWrapper"
      endBg.id = "movingBg"
      document.body.appendChild(endWrapper)
      endWrapper.appendChild(endBg)
      endWrapper.appendChild(endScreen)
      titlemusic = new Audio(musicData["title"])
      playMusic(titlemusic)
      rollCredits(endWrapper, [restartButton])
    }
  }

//=========================================================
//INITIALIZATION of DOM

class DOMDisplay {
  constructor(parent, level) {
    this.dom = elt("div", {class: "game"}, drawGrid(level));
    this.actorLayer = null;
    this.dialoguebox = document.createElement("div");
    parent.appendChild(this.dom);
  }

  clear() { this.dom.remove(); }

  hide() {
    this.dom.setAttribute("hidden", "");
  }

  show () {
    this.dom.removeAttribute("hidden");
  }

  deleteDialogue() {
    this.dialoguebox.remove();
    let dialogues = document.querySelectorAll(".container")
    if (dialogues) {
      for (let i = 0; i < dialogues.length; i++) {
      dialogues[i].remove()
      }
    }
    
  }

  createDialogue(dom, status, data) {

    this.dialoguebox = document.createElement("div")

    printDialogueData(this.dialoguebox, 0, dom, this.dialoguebox, status, data)
    
  }
  
}

//=========================================================
//Function to create ingame tutorial
function createTutorial(dom) {
    tutorialbox = document.createElement("div")
    tutorialbox.id = "tutorial"
    document.body.appendChild(tutorialbox)
    return tutorialbox
  }

//=========================================================
//DOM functions
DOMDisplay.prototype.syncState = function(state) {
  if (this.actorLayer) this.actorLayer.remove();
  this.actorLayer = drawActors(state.actors);
  this.dom.appendChild(this.actorLayer);
  this.dom.className = `game ${state.status}`;
  this.scrollPlayerIntoView(state);
};

DOMDisplay.prototype.scrollPlayerIntoView = function(state) {
  if (state.level.type == "level") return
  let player = state.player
  let width = this.dom.clientWidth - 15;
  let height = this.dom.clientHeight - 15;
  let center = player.pos.plus(player.size.times(0.5)).times(scale)
  //=========================================================
  //Direct scrolling
  let start = new Vec(center.x / state.level.width, center.y / state.level.height)
  let end = new Vec(start.x + width / state.level.width, start.y + height / state.level.height)
  this.dom.scrollLeft = center.x - (width / 2)
  this.dom.scrollTop = center.y - (height / 2)

  //=========================================================
  //Only scroll on edge of screen code:
  /*
  let width = this.dom.clientWidth - 15;
  let height = this.dom.clientHeight - 15;
  let margin = width / 5;

  // The viewport
  let left = this.dom.scrollLeft, right = left + width;
  let top = this.dom.scrollTop, bottom = top + height;

  let player = state.player;
  let center = player.pos.plus(player.size.times(0.5))
                         .times(scale);

  if (center.x < left + margin) {
    this.dom.scrollLeft = center.x - margin;
  } else if (center.x > right - margin) {
    this.dom.scrollLeft = center.x + margin - width;
  }
  if (center.y < top + margin) {
    this.dom.scrollTop = center.y - margin;
  } else if (center.y > bottom - margin) {
    this.dom.scrollTop = center.y + margin - height;
  }
  */
};

//=========================================================
//Function to transform data used in Eloquent Javascript Tutorial
function elt(name, attrs, ...children) {
  let dom = document.createElement(name);
  for (let attr of Object.keys(attrs)) {
    dom.setAttribute(attr, attrs[attr]);
  }
  for (let child of children) {
    dom.appendChild(child);
  }
  return dom;
}


//=========================================================
//Creates a Grid to display the game
function drawGrid(level) {
  scalewidth = (window.innerWidth - 15)/level.width;
  scaleheight = (window.innerHeight - 15)/level.height;
  if (scalewidth > scaleheight) {
    scale = scaleheight;
  }
  else {
    scale = scalewidth;
  }
  scale = scalewidth
  if (level.type == "map") {
    scale = 50;
  }
  return elt("table", {
    class: "background",
    style: `width: ${level.width * scale}px; background: ${levelplan[currentlevel]["enemy"] == "Rufus" ? "black" : "rgb(52, 166, 251)"}`
  }, ...level.rows.map(row =>
    elt("tr", {style: `height: ${scale}px`},
        ...row.map(type => elt("td", {class: `${type}`})))
  ));
}

//=========================================================
//Adds Actors: Moving objects
function drawActors(actors) {
  return elt("div", {}, ...actors.map(actor => {
    let rect = elt("div", {class: `actor ${actor.type}`});
    // rect.style.backgroundImage = levelplan[currentlevel]["enemy"] == "Rufus" && actor.type == "event" ? "url(./Textures/Background.png)" : rect.style.backgroundImage
    rect.style.filter = levelplan[currentlevel]["enemy"] == "Rufus" && actor.type == "event" ? "opacity(0%)" : rect.style.opacity
    rect.style.width = `${actor.size.x * scale}px`;
    rect.style.height = `${actor.size.y * scale}px`;
    rect.style.left = `${actor.pos.x * scale}px`;
    rect.style.top = `${actor.pos.y * scale}px`;
    rect.style.transform = arrowKeys.ArrowRight && actor.type == "player" ? "scaleX(1)" : `scaleX(${actor.dir})`
    rect.style.transform = arrowKeys.ArrowLeft && actor.type == "player" ? "scaleX(-1)" : `scaleX(${actor.dir})`
    rect.style.transform = actor.type.includes("attack") ? "scaleX(1)" : `scaleX(${actor.dir})`
    let sprite = document.createElement("div")
    sprite.classList.add("sprite")
    sprite.style.height = rect.style.height;
    sprite.style.width = rect.style.height;
    sprite.style.backgroundImage = actor.type == "player" ? `url("${walkingSpriteData["Aysel"][Math.round(actor.frame)]}")` : sprite.style.backgroundImage
    sprite.style.backgroundImage = actor.type == "enemy" ? `url("${levelplan[currentlevel]["enemySprites"][Math.round(actor.frame)]}")` : sprite.style.backgroundImage
    sprite.style.backgroundPositionY = actor.type == "player" && actor.frame < jumpFrame ? `${scale * 0.3}px`: sprite.style.backgroundPositionY;
    sprite.style.backgroundPositionY = actor.type == "enemy" && levelplan[currentlevel]["enemySprites"] == walkingSpriteData["Lindor"] && actor.frame >= 0 && actor.frame <= 2 ? `${scale * 0.1}px`: sprite.style.backgroundPositionY;
    sprite.style.backgroundPositionY = actor.type == "enemy" && levelplan[currentlevel]["enemySprites"] == walkingSpriteData["Rufus"] && actor.frame >= 0 && actor.frame <= 2 ? `${scale * 0.1}px`: sprite.style.backgroundPositionY;
    sprite.style.backgroundPositionY = actor.type == "enemy" && levelplan[currentlevel]["enemySprites"] == walkingSpriteData["Petros"] && actor.frame >= 0 && actor.frame <= 2 ? `${scale * 0.1}px`: sprite.style.backgroundPositionY;

    sprite.style.backgroundPositionY = actor.type == "enemy" && levelplan[currentlevel]["enemySprites"] == walkingSpriteData["Sporky"] && actor.frame >= 0 && actor.frame <= 2 ? `${scale * 0.1}px`: sprite.style.backgroundPositionY;
    if (actor.type == "player") rect.appendChild(sprite)
    if (actor.type == "enemy") rect.appendChild(sprite)

    rect.style.transform = actor.type.includes("bounce") ? `rotate(${actor.speed.y * actors.find(a => a.type == "enemy").timer}deg)` : rect.style.transform
rect.style.transform = actor.type.includes("rotatey") ? `rotate(${actor.angle}deg)` : rect.style.transform
    rect.style.backgroundPositionY = actor.type == "powerbar" ?   `${-actor.shift / 2}px` : rect.style.backgroundPositionY
    rect.style.backgroundPositionX = actor.type == "hp" ? `${actor.shift / 2}px` : rect.style.backgroundPositionX
    // rect.style.fontSize = actor.type == "hp" ? `${rect.style.offsetHeight}px` : rect.style.fontSize
    rect.innerHTML = actor.type == "hp" ? (actor.char == "player" ? globalPlayerHP.toFixed(1) : globalEnemyHP.toFixed(1)) : rect.innerHTML
    return rect;
  }));
}

//=========================================================
//Returns if specific pos touches an object of the type input
Level.prototype.touches = function(pos, size, type) {
  let xStart = Math.floor(pos.x);
  let xEnd = Math.ceil(pos.x + size.x);
  let yStart = Math.floor(pos.y);
  let yEnd = Math.ceil(pos.y + size.y);

  for (let y = yStart; y < yEnd; y++) {
    for (let x = xStart; x < xEnd; x++) {
      let isOutside = x < 0 || x >= this.width ||
                      y < 0 || y >= this.height;
      let here = isOutside ? "wall" : this.rows[y][x];
      if (here == type) return true;
    }
  }
  return false;
};

//=========================================================
//Main loop of the game, updates everything in the Level and Map
State.prototype.update = function(time, keys) {
  let actors = this.actors.map(actor => actor.update(time, this, keys));
  let newState = new State(this.level, actors, this.status);
  if (arrowKeys[" "] && newState.status == "reading") {
    arrowKeys = mapKeys
    newState.status = "playing";
  }

  if (newState.status != "playing") return newState;

  let player = newState.player;

  for (let actor of actors) {
    if (actor.type == "projectile fast spork") {
      if (actor.timer > 4 * Math.PI) {
        actors.splice(actors.indexOf(actor), 1)
      }
    }
    if (actor != player && overlap(actor, player)) {
      newState = actor.collide(newState);
    }
    if (actor.pos.x < 0 && actor.type.includes("attack")) {
      if (actor.dir != -1) {
        globalPlayerHP -= 1;
      actors.splice(actors.indexOf(actor), 1)
      playSoundEffect(damage)
      if (globalPlayerHP < 1) {
        if (newState.tutorial != "") newState.tutorial[0].remove()
        return new State(this.level, actors, "lost")
      }
      }
    }
    if (actor.pos.x > newState.level.width && actor.type.includes("attack")) {
      if (actor.dir == -1) {
        globalPlayerHP -= 1;
        actors.splice(actors.indexOf(actor), 1)
        playSoundEffect(damage)
        if (globalPlayerHP < 1) {
          if (newState.tutorial != "") newState.tutorial[0].remove()
          return new State(this.level, actors, "lost")
        }
      }
    }
    if (actor.pos.y > newState.level.height && actor.type.includes("attack") && actor.type.includes("line") && attackTypes.includes(actor.type.split(" ")[1])) {
      if (actor.dir != -1)
      {
        globalPlayerHP -= 1;
        actors.splice(actors.indexOf(actor), 1)
        playSoundEffect(damage)
        if (globalPlayerHP < 1) {
          if (newState.tutorial != "") newState.tutorial[0].remove()
          return new State(this.level, actors, "lost")
        }
      }
    }
    if (actor.pos.y < 0 && actor.type.includes("attack") && actor.type.includes("line") && attackTypes.includes(actor.type.split(" ")[1])) {
      if (actor.dir == -1)
      {
        globalPlayerHP -= 1;
        actors.splice(actors.indexOf(actor), 1)
        playSoundEffect(damage)
        if (globalPlayerHP < 1) {
          if (newState.tutorial != "") newState.tutorial[0].remove()
          return new State(this.level, actors, "lost")
        }
      }
    }
    if (actor.type == "enemy") {
      //if (levelplan[currentlevel]["leveltype"] == "normal") {
      newState.tutorial = this.tutorial
      if (actor.timer % 100 == 0 && newState.level.touches(actor.pos.plus(new Vec(0, .05)), actor.size, "wall")) {
        createAttack(newState)
      }
      else if (Math.round(actor.frame) == 0 && Math.floor(Math.random() * 100) == 0) {
        actor.frame = eJumpFrame + .4;
      }
      //}
      if (levelplan[currentlevel]["leveltype"] == "tutorial") {
        newState.tutorial = [createTutorial(), 0]
        levelplan[currentlevel]["leveltype"] = "tutorial(inprogress)"
      }
      else if (levelplan[currentlevel]["leveltype"] == "tutorial(inprogress)") {
        newState.tutorial[0].innerHTML = tutorialText[newState.tutorial[1]]
        newState.tutorial[0].style.fontSize = newState.tutorial[0].offsetHeight / 10
        if (newState.tutorial[1] == 0) {
          attackData = []
          if (tutorialtimer == true) {
            setTimeout(function () {
            newState.tutorial[1] ++
            tutorialtimer = true
          }, 10000)
          }
          tutorialtimer = false;
          
        }
        else if (newState.tutorial[1] == 1) {
          attackData = [Line]
        }
        else if (newState.tutorial[1] == 3) {
          attackData = [RockAttack]
        }
        else if (newState.tutorial[1] == 5) {
          attackData = [RockAttack, Line]
        }
        else if (newState.tutorial[1] == 8) {
          attackData = []
          if (tutorialtimer == true) {
            setTimeout(function () {
            newState.tutorial[0].remove()
            tutorialtimer = "finish";
          }, 10000)
          }
          if (tutorialtimer == "finish") return new State(this.level, actors, "won")
          if (tutorialtimer != "finish") tutorialtimer = false
        }
      }

    }
    if (actor.type == "powerbar") {
      if (attackDirection == "LR" && power > 0) {
        let newCounterattack = CounterAttackH.create(newState.player.pos, power)
        newState.actors.push(newCounterattack)
        playSoundEffect(counterattack)
        power = 0
      }
      else if (attackDirection == "UD" && power > 0) {
        let newCounterattack = CounterAttackV.create(new Vec(newState.level.width - newState.player.pos.x - 1, -1), power)
        newState.actors.push(newCounterattack)
        playSoundEffect(counterattack)
        power = 0
      }
    }
    if (actor.pos.x < 0) { //Horizontal Projectile
      if (actor.type.includes("projectile") && levelplan[currentlevel]["leveltype"] == "tutorial(inprogress)"){
        if (newState.tutorial[1] == 1 || newState.tutorial[1] == 2) {
          newState.tutorial[1] ++
        }
        actors.splice(newState.actors.indexOf(actor), 1)
      }
    }
    if (actor.pos.y > newState.level.height && actor.type.includes("line") && actor.type.includes("projectile") && levelplan[currentlevel]["leveltype"] == "tutorial(inprogress)") { //Vertical Projectile
    if (newState.tutorial[1] == 1 || newState.tutorial[1] == 2) {
        newState.tutorial[1] ++
      }
      actors.splice(newState.actors.indexOf(actor), 1)
    }
    if (actor.type.includes("counterattack")) {
      if (actor.pos.x > newState.level.width) {
        newState.actors.splice(newState.actors.indexOf(actor), 1)
      }
      else if (actor.pos.y > newState.level.height) {
        newState.actors.splice(newState.actors.indexOf(actor), 1)
      }
      if(overlap(actor, newState.enemy)) {
        newState.actors.splice(newState.actors.indexOf(actor), 1)
        if (actor.type.includes("horizontal"))
        {
          if (levelplan[currentlevel]["leveltype"] == "tutorial(inprogress)") {
            if (newState.tutorial[1] == 6) {
              newState.tutorial[1] ++
            }
          }
          globalEnemyHP -= (actor.speed.x + actor.speed.y) / 10
        }
        else if (actor.type.includes("vertical")) {
          if (levelplan[currentlevel]["leveltype"] == "tutorial(inprogress)") {
            if (newState.tutorial[1] == 7) {
              newState.tutorial[1] ++
            }
          }
          globalEnemyHP -= (actor.speed.x + actor.speed.y) / 5
        }
        playSoundEffect(damage)
        if (globalEnemyHP <= 0) {
          if (newState.tutorial != "") newState.tutorial[0].remove()
          return new State(this.level, actors, "won")
        }
      }
    }
  }
  return newState;
};

//=========================================================
//Returns if two actors overlap
function overlap(actor1, actor2) {
  return actor1.pos.x + actor1.size.x > actor2.pos.x &&
         actor1.pos.x < actor2.pos.x + actor2.size.x &&
         actor1.pos.y + actor1.size.y > actor2.pos.y &&
         actor1.pos.y < actor2.pos.y + actor2.size.y;
}

//=========================================================
//Function to regulate blocking of Attacks
function attackCollision(state, attack) {
  if (attackDirection.includes(attack.ch)) {
    if (levelplan[currentlevel]["leveltype"] == "tutorial(inprogress)") {
      if ([3, 4, 5].includes(state.tutorial[1])) {
        state.tutorial[1] ++
      }
    }
    state.actors.splice(state.actors.indexOf(attack), 1)
    if (power >= 10) power = 10
    else {
      power += 1;
      if (power == 10) {
        playSoundEffect(maxpower)
      }
    }
    
  }
  return state;
}

function projectileCollision(state, attack) {
  globalPlayerHP --;
  state.actors.splice(state.actors.indexOf(attack), 1)
  if (globalPlayerHP < 1) {
    state.status = "lost"
    if (state.tutorial != "")       state.tutorial[0].remove()
  }
  playSoundEffect(damage)
  return state;
}

//=========================================================
//All Actor collisions
Attack.prototype.collide = function(state) {
  return attackCollision(state, this);
}

CandentisAttack.prototype.collide = function(state) {
  return attackCollision(state, this);
}

CandentisVAttack.prototype.collide = function(state) {
  return attackCollision(state, this);
}

PizzaAttack.prototype.collide = function(state) {
  return attackCollision(state, this);
}

RockAttack.prototype.collide = function(state) {
  return attackCollision(state, this);
}

CounterAttackH.prototype.collide = function(state) {
  return state;
}

CounterAttackV.prototype.collide = function(state) {
  return state;
}

SineAttack.prototype.collide = function (state) {
  return attackCollision(state, this);
}

BounceProjectile.prototype.collide = function(state) {
  return projectileCollision(state, this)
  // globalPlayerHP --;
  // state.actors.splice(state.actors.indexOf(this), 1)
  // if (globalPlayerHP < 1) {
  //   state.status = "lost"
  //   if (state.tutorial != "") state.tutorial[0].remove()
  // }
  // playSoundEffect(damage)
  // return state;
}

SporkProjectile.prototype.collide = function(state) {
  return projectileCollision(state, this)
}

Line.prototype.collide = function(state) {
  return projectileCollision(state, this)
}

BigSpork.prototype.collide = function(state) {
  return projectileCollision(state, this)
}

Dialogue.prototype.collide = function(state) {
  if (arrowKeys["e"]) {
    arrowKeys = []
    state.player.pos.x = this.pos.x
    state.player.pos.y = this.pos.y - 1.5/2
    return new State(state.level, state.actors, "dialoguestart")
  }
  else {
    return state;
  }
}

Stairs.prototype.collide = function(state) {
  arrowKeys = []
  playSoundEffect(stairssound)
  return new State(state.level, state.actors, "goup")
}

Event.prototype.collide = function(state) {
  if (levelplan[currentlevel]["enemy"] != "Rufus") arrowKeys = []
  return new State(state.level, state.actors, "levelstart")
}


//=========================================================
//All actor update functions
Attack.prototype.update = function(time, state) {
  let newPos = this.pos.plus(this.speed.times(time));
  if (newPos.x > 0) {
    return new Attack(newPos, this.speed, this.ch)
  }
  else {
    return new Attack(newPos, this.speed, this.ch)
  }
}

CandentisAttack.prototype.update = function(time, state) {
  //this.dir == -1 ? this.type = `attack ${this.ch} candentis backwards` : true
  let newPos = this.pos.plus(this.speed.times(time));
  this.speed = new Vec(Math.random() * -6 * this.dir + 1, 0)
  this.pos = newPos
  if (newPos.x > 0) {
    return this
  }
  else {
    return this
  }
}

CandentisVAttack.prototype.update = function(time, state) {
  let newPos = this.pos.plus(this.speed.times(time));
  this.speed = new Vec(0, Math.random() * 4 * this.dir)
  this.pos = newPos
  if (newPos.x > 0) {
    return this
  }
  else {
    return this
  }
}

PizzaAttack.prototype.update = function(time, state) {
  let newPos = this.pos.plus(this.speed.times(time));
  this.speed.x -= .02
  if (newPos.x > 0) {
    return new PizzaAttack(newPos, this.speed, this.ch)
  }
  else {
    return new PizzaAttack(newPos, this.speed, this.ch)
  }
}

RockAttack.prototype.update = function(time, state) {
  let newPos = this.pos.plus(this.speed.times(time));
  if (newPos.x > 0) {
    return new RockAttack(newPos, this.speed, this.ch)
  }
  else {
    return new RockAttack(newPos, this.speed, this.ch)
  }
}

CounterAttackH.prototype.update = function(time, state) {
  let newPos = this.pos.plus(this.speed.times(time))
  return new CounterAttackH(newPos, this.speed)
}

CounterAttackV.prototype.update = function(time, state) {
  let newPos = this.pos.plus(this.speed.times(time))
  return new CounterAttackV(newPos, this.speed)
}

SineAttack.prototype.update = function(time, state) {
  let newPos = new Vec(this.speed.x * time + this.pos.x, this.pos.y + Math.sin(this.pos.x) * time * this.wobble)
  return new SineAttack(newPos,this.speed, this.ch, this.wobble)
}

BounceProjectile.prototype.update = function(time, state) {
  let newPos = new Vec(this.speed.x * time + this.pos.x, this.speed.y * time + this.pos.y)
  this.pos = newPos;
  if (newPos.y <= 0 || newPos.y >= state.level.height) {
    this.speed.y = this.speed.y * -1;//fix
  }
  return this;
}

SporkProjectile.prototype.update = function(time, state) {
  let newPos = new Vec(this.speed.x * time + this.pos.x, this.speed.y * time + this.pos.y)
  this.pos = newPos;
  this.speed.y = this.speed.y + 90 * (time**2)
  this.angle = Math.atan(this.speed.y / this.speed.x) * 180 / Math.PI 
  return this;
}

BigSpork.prototype.update = function(time, state) {
  if (this.timer < 2 * Math.PI) {
    this.timer += 0.05
    this.speed = new Vec(0, 0)
  }
  let newPos = new Vec(this.speed.x * time + this.pos.x, this.speed.y * time + this.pos.y)
  this.pos = newPos;
  this.timer += 0.05
  this.speed.y = this.pos.y > 5 ? Math.sin(this.timer + Math.PI) * 5 : Math.sin(this.timer) * 5
  return this;
}

Line.prototype.update = function(time, state) {
  let newPos = new Vec(this.speed.x * time + this.pos.x, this.speed.y * time + this.pos.y)
  this.pos = newPos;
  return this;
}

const playerXSpeed = 7;
const gravity = 30;
const jumpSpeed = 17;

Dialogue.prototype.update = function() {
  return this;
}

Stairs.prototype.update = function() {
  return this;
}

HP.prototype.update = function() {
  this.value = this.char == "player" ? new Vec(globalPlayerHP,1): new Vec(globalEnemyHP * 5 / levelplan[currentlevel]["enemyHealth"], 1);
  this.size = this.value;
  this.shift ++
  return this;
}

PowerLevel.prototype.update = function() {
  this.size = levelplan[currentlevel]["enemy"] != "Rufus" ? new Vec(1, power) : new Vec(power, 1)
  this.pos = levelplan[currentlevel]["enemy"] != "Rufus" ? new Vec(2, 11 - power) : new Vec(9.5 - power / 2, 1)
  this.shift ++
  return this;
}

Event.prototype.update = function() {
  return new Event(this.pos);
}

Enemy.prototype.update = function(time, state, keys) {
  let ySpeed = this.speed.y + time * gravity;
  let movedY = this.pos.plus(new Vec(0, ySpeed * time));
  if (!state.level.touches(movedY, this.size, "wall")) {
    this.pos = movedY;
  } 
  else if (Math.round(this.frame) == eJumpFrame && this.frame < eJumpFrame && Math.floor(Math.random() * 5) == 0) {
      playSoundEffect(jump)
      ySpeed -= Math.floor(Math.random() * jumpSpeed);
  } 
  else {
    ySpeed = 0;
  }
  if (Math.round(this.frame) == eJumpFrame) {
    this.frame -= 0.1;
    if (this.direction == 0) {
      this.direction = (Math.floor(Math.random() * 3 - 1)) / 10
    }
    newPos = this.pos.plus(new Vec(this.direction * this.frame / 5, 0))
    if (!state.level.touches(newPos, this.size, "wall")) {
      this.pos = newPos
    }
    if (Math.round(this.frame) != eJumpFrame) {
      this.frame = 0;
      this.direction = 0;
    }
  }
  else if (Math.round(this.frame) > 0 && Math.floor(this.frame) <= 2) {
    this.frame -= 0.1;
    if (this.frame <= 0) {
      this.frame = 0;
    }
  }

  newPos = this.pos.plus(new Vec(0, ySpeed * time))
    
  if (!state.level.touches(newPos, this.size, "wall")){
    this.pos = newPos
  }

  if (!state.level.touches(this.pos.plus(new Vec(0, .1)), this.size, "wall")) {
      this.frame = eJumpFrame
    }

  this.speed.y = ySpeed
  this.timer ++;
  return this;
}

Player.prototype.update = function(time, state, keys) {
  attackDirection = "";
  let xSpeed = 0;
  if (keys.ArrowLeft) {
    xSpeed -= playerXSpeed;
    this.dir = -1
  }
  else if(keys.ArrowRight) {
    xSpeed += playerXSpeed;
    this.dir = 1
  }
  if (state.level.leveltype == "level") {
    if (levelKeys.w) attackDirection = `${attackDirection}U`
    if (levelKeys.a) attackDirection = `${attackDirection}L`
    if (levelKeys.s) attackDirection = `${attackDirection}D`
    if (levelKeys.d) attackDirection = `${attackDirection}R`

    if (attackDirection != "") playSoundEffect(attack)

    if (levelKeys.d) {
      xSpeed += playerXSpeed * 5;
    }
    if (levelKeys.a) {
      xSpeed -= playerXSpeed * 5;
    }
  }
  let pos = this.pos;
  let movedX = pos.plus(new Vec(xSpeed * time, 0));
  if (!state.level.touches(movedX, this.size, "wall")) {
    pos = movedX;
  }
  let ySpeed = 0;
  if (state.level.leveltype == "map")
  {
    if (!keys.ArrowLeft && !keys.ArrowRight && !keys.ArrowUp && !keys.ArrowDown) {
    this.frame = 0
    }

    if (keys.ArrowUp || keys.ArrowDown || keys.ArrowRight || keys.ArrowLeft) {
          this.frame = this.frame < walkingFrames ? this.frame += .1 : 1
    }

    if (keys.ArrowUp) {
      ySpeed -= playerXSpeed;
    }
    if (keys.ArrowDown) {
      ySpeed += playerXSpeed;
    }
    let movedY = pos.plus(new Vec(0, ySpeed * time));
    if (!state.level.touches(movedY, this.size, "wall")) {
      pos = movedY;
    }
  }
  else {
    ySpeed = this.speed.y + time * gravity;
    if (levelKeys.s) {
      ySpeed += playerXSpeed * 2;
    }
    if (levelKeys.w && this.doubleJump) {
      ySpeed -= jumpSpeed / 2;
      this.doubleJump = false;
      playSoundEffect(doublejump)
    }
    let movedY = pos.plus(new Vec(0, ySpeed * time));
    if (!state.level.touches(movedY, this.size, "wall")) {
      pos = movedY;
      if (!keys.ArrowUp) {
        this.frame = fallFrame;
      }
    } 
    else if (keys.ArrowUp && ySpeed > 0) {
      playSoundEffect(jump)
      ySpeed = -jumpSpeed;
      this.frame = jumpFrame;
      
    } 
    else {
      ySpeed = 0;
      this.doubleJump = true;
      if (keys.ArrowLeft || keys.ArrowRight) {
        this.frame = this.frame < walkingFrames ? this.frame += .1 : 1
      }
      if (!keys.ArrowLeft && !keys.ArrowRight && !keys.ArrowUp && !keys.ArrowDown) {
        this.frame = 0
      }
    }

    if (levelKeys.w) {
      this.frame = upFrame;
    }
    if (levelKeys.a) {
      this.frame = leftrightFrame;
      this.dir = -1
    }
    if (levelKeys.s) {
      this.frame = downFrame;
    }
    if (levelKeys.d) {
      this.frame = leftrightFrame;
      this.dir = 1
    }

  }

  return new Player(pos, new Vec(xSpeed, ySpeed), this.frame, this.dir, this.doubleJump);
};

//=========================================================
//Function that checks what keys are pressed, can be specified to keydown or keyup events
function trackKeys(keys, eventList) {
  let down = Object.create(null);
  function track(event) {
    if (keys.includes(event.key)) {
      down[event.key] = event.type == "keydown";
      event.preventDefault();
      if (eventList[keys.indexOf(event.key)] == "keyup" && event.repeat) {
        down[event.key] = false;
      }
    }
  }
  window.addEventListener("keydown", track);
  window.addEventListener("keyup", track);
  return down;
}

//=========================================================
//Runs the animation using requestAnimationFrame and a recursive function
function runAnimation(frameFunc) {
  let lastTime = null;
  function frame(time) {
    if (lastTime != null) {
      let timeStep = 60 / 3500 //Math.min(time - lastTime, 100) / 1000;
      if (frameFunc(timeStep) === false) return;
    }
    lastTime = time;
    setTimeout(requestAnimationFrame(frame), 60, frame);
  }
  requestAnimationFrame(frame);
}

//=========================================================
//Functions used for placing Enemy Attacks
function randomRightPos(state) {
  return new Vec(state.level.width, (Math.floor(Math.random() * 7)) + 4)
}

function randomLeftPos() {
  return new Vec(-1, (Math.floor(Math.random() * 7)) + 4)
}

function randomTopPos() {
  return new Vec((Math.floor(Math.random() * 12)) + 3, -1)
}

function randomBottomPos(state) {
  return new Vec((Math.floor(Math.random() * 12)) + 3, state.level.height)
}

function findStartPos(attackClass, state, tobedir = false) {
  let classy = new attackClass
    if (attackClass == Line || attackClass == RockAttack) {
      return randomTopPos()
    }
    else if (attackClass == BigSpork) {
      let directions = [-1, 1]
      let dir = directions[Math.floor(Math.random() * directions.length)]
      return dir > 0 ? new Vec(5, 13) : new Vec(5, -9)
    }
      else if (classy.dir) {
        if (tobedir == -1) {
          if (attackClass == CandentisAttack)
          {
            return randomLeftPos()
          }
          else if (attackClass == CandentisVAttack) {
            return randomBottomPos(state)
          }
        }
        else if (tobedir == 1) {
          if (attackClass == CandentisAttack) {
            return randomRightPos(state)
          }
          else if (attackClass == CandentisVAttack) {
            return randomTopPos()
          }
          
        }
      }
    else {
      return randomRightPos(state)
    }
}

//=========================================================
//Creates an Enemy's attack object
function createAttack(state) {
  if (attackData.length == 0) return
  if (state.status == "playing" && Math.floor(Math.random() * 2)) {
    let attackClass = attackData[Math.floor(Math.random() * attackData.length)]
    let tobedir = Math.floor(Math.random() * 2) == 1 ? 1 : -1
    let startingPos = findStartPos(attackClass, state, tobedir)
    let newAttack = attackClass.create(startingPos, attackTypes[Math.floor(Math.random() * attackTypes.length)]);
          state.enemy.frame = 2.4;
    if (newAttack.warning) {
      warning.play()
    }
    else if (newAttack.dir && attackClass != BigSpork) {
      newAttack.dir = tobedir
      playSoundEffect(tennisball)
    }
    else {
      playSoundEffect(tennisball)
    }
    state.actors.push(newAttack)

    if (Math.floor(Math.random() * 2)) {
      attackClass = attackData[Math.floor(Math.random() * attackData.length)]
    startingPos = findStartPos(attackClass, state, tobedir)
      newAttack = attackClass.create(startingPos, attackTypes[Math.floor(Math.random() * attackTypes.length)]);
    if (newAttack.warning) {
      warning.play()
    }
    else if (newAttack.dir && attackClass != BigSpork) {
      newAttack.dir = tobedir
      playSoundEffect(tennisball)
    }
    state.actors.push(newAttack)
    }
  }
}

//=========================================================
//Runs either a map or level
function runLevel(level, Display) {
power = 0
globalEnemyHP = levelplan[currentlevel]
  ["enemyHealth"]
  let ending = 1;
  let state = State.start(level);
  if (state.level.leveltype == "map") playMapMusic()
   let display = new Display(document.body, level);
  if (state.level.leveltype == "level") {
        let playerHP = levelplan[currentlevel]["enemy"] != "Rufus" ? new HP(new Vec(3, 11), "player"): new HP(new Vec(7, 11), "player")
        let enemyHP = new HP(new Vec(30,11), "enemy")
        let powerLevel = levelplan[currentlevel]["enemy"] != "Rufus" ? new PowerLevel(new Vec(2,11)) : new PowerLevel(new Vec(9.5, 1))
        state.actors.push(playerHP)
        state.actors.push(enemyHP)
        state.actors.push(powerLevel)
        musicData["level"] = levelplan[currentlevel]["music"]
        bgMusic = new Audio(musicData["level"])
        playMusic(bgMusic)
  }
         return new Promise(resolve => {
          runAnimation(time => {
          if (state.status == "lost") {
        bgMusic.pause()
        display.clear();
        resolve(state.status)
        return true;
      }
      if (state.status == "playing") {
        display.deleteDialogue();
      }

      state = state.update(time, arrowKeys);
      display.syncState(state);
      if (state.status == "playing") {
        display.deleteDialogue()
        display.show()
        return true;
      } else if (ending > 0) {
        ending -= time;
        return true;
      } else if (state.status == "won") {
        display.clear();
        bgMusic.pause();
        resolve(state.status);
        state.status = "gamewin"
        arrowKeys = dialogueKeys
        display.createDialogue(this.dom, state.status, levelplan[currentlevel]["windata"])
        return false;
      }
      else if (state.status == "lost") {
        bgMusic.pause()
        display.clear();
        resolve(state.status)
        state.clear()
        return true;
      }
      else if (state.status == "dialoguestart") {
        display.hide();
        let data = levelplan[currentlevel]["dialoguegrid"][state.player.pos.x][state.player.pos.y+1.5/2]
//=========================================================
        //Code below calculates distance but changed to direct player position approach (less time and works better):
        /*
        let dist = new Vec(state.level.width, state.level.height)
        let index = 0
        for (let i = 0; i < data.length; i++) {
          let thisdist = new Vec(Math.abs(data[i].pos.x - state.player.pos.x), Math.abs(data[i].pos.y - state.player.pos.y))
          if (thisdist.x + thisdist.y) < (dist.x + dist.y)) {
            dist = thisdist
            index = i
          }
        }
        data = data[index].data
          */
        state.status = "reading"
        arrowKeys = dialogueKeys
        let dialoguebox = display.createDialogue(this.dom, state.status, data);
        return true
      }
      else if (state.status == "levelstart") {
        display.clear();
        resolve(state.status)
        return false;
      }
      else if (state.status == "goup") {
        display.clear();
        resolve(state.status)
        return false;
      }
        });
        });
  }

//=========================================================
//Function to make music fade out and pause
async function fadeOut(music) {
  for (i = 1; i > 0; i-= 0.001) {
    music.volume = i
  }
  music.pause();
}

//=========================================================
//Runs a level but recursively used to run the entire game
//A game over reloads the page
async function runGame(plans, Display) {
  let status = await runLevel(plans[0], Display)
  if (status == "goup") {
    currentlevel ++
    let testMap = new Level(levelplan[currentlevel]["map"], "map"); //used to be map
    let testLevel = new Level(levelplan[currentlevel]["level"], "level");
    arrowKeys = mapKeys
    createtextdesc(testMap, testLevel, DOMDisplay)
        
  }
  if (status == "levelstart") {
    pauseMapMusic()
    let banner = document.createElement("div")
    let text = document.createElement("div")
    banner.id = "banner"
    banner.style.backgroundImage = `url("${levelplan[currentlevel]["enemybanner"]}")`
    text.id = "bannercaption"
    text.innerHTML = levelplan[currentlevel]["enemy"]
    banner.appendChild(text)
    bannermusic = new Audio(musicData["banner"])
    playMusic(bannermusic)
    playSoundEffect(enemybanner)
    document.body.appendChild(banner)
    setTimeout(async function() {
      banner.remove()
      arrowKeys = levelKeys
      fadeOut(bannermusic)


      let levelstatus = await runLevel(plans[1], Display)
      if (levelstatus == "lost") {
        let gameOverMusic = new Audio(musicData["gameover"])
        let gameOverScreen = document.createElement("div")
        gameOverScreen.id = "gameover"
        let restartButton = document.createElement("button")
        restartButton.id = "restartbutton"
        restartButton.innerHTML = "Restart"
        gameOverScreen.appendChild(restartButton)
        document.body.appendChild(gameOverScreen)
        playMusic(gameOverMusic)
        restartButton.onclick = function() {
          location.reload()
        }
      
    }

    }, 5000)

  }
}